(function() {
	'use strict';

	process.title = 'Hojta';
	var http = require('http');
	var spawn = require('child_process').spawn;
	var exec = require('child_process').exec;
	var fs = require('fs');
	var dive = require('dive');
	var async = require('async');
	var extend = require('extend');
	var path = require('path');

	// Check if app is in debug mode or not.
	var debugMode = false;
	try {
		fs.statSync(__dirname + '\\..\\debug.txt');
		debugMode = true;
	} catch (e) {
		console.log(e);
	}

	var hangingRequests = {};

	var lastErrorPath = path.resolve(__dirname + '\\..\\last_error.txt');

	process.on('uncaughtException', function(err) {
		var errorLogPath = path.resolve(__dirname + '\\..\\error.log.txt');
		var version = fs.readFileSync(__dirname + '\\..\\version').toString();

		// We don't want to crash if someone makes a broken plugin...
		// But make sure to log it
		var report = ['Hojta (', version, ') - ', new Date(), '\n',
			err.stack, '\n\n---\n\n'
		].join('');

		report = report.replace(/\n/g, '\r\n');

		// Finalize all hanging requests, yes this is bad... But it should
		// never have to come to this. Better to fix the plugin.
		Object.keys(hangingRequests).forEach(function(id) {
			var httpRes = hangingRequests[id];
			httpRes.end(JSON.stringify({
				'textValues': {},
				'conditions': {}
			}));
		});
		hangingRequests = {};

		fs.writeFile(errorLogPath, report, {
			'flag': 'a'
		}, function() {
			if (debugMode) {
				return console.log(report);
			}

			// If we aren't in debug mode and an error occurs show that to the
			// user so they can give a proper report to the plugin creator.
			fs.writeFile(lastErrorPath, report, function() {
				exec('notepad ' + lastErrorPath, function() {
					fs.unlink(lastErrorPath, function() {
						process.exit(0);
					});
				});
			});
		});
	});

	// Cleanup
	fs.unlink(lastErrorPath, function() {});

	var helpersGenerator = (function() {
		// Takes mp3 (only) file path as arg.
		// Can be relative from sounds folder as well as plugin folder.
		function playSound(pluginOptions) {
			var bin = '..\\bin\\PlaySound.exe';

			// Allow people to put files within the plugin folders
			function check(soundPaths) {
				var file = soundPaths[0];
				if (!file) {
					console.log('Could not play ' + fileName);
					return;
				}

				soundPaths = soundPaths.slice(1);

				function validateFile(err) {

					if (err) {
						return check(soundPaths);
					}

					spawn(bin, [file]);
				}

				fs.stat(file, validateFile);
			}

			return function(fileName) {
				// Check different folders for sound
				var soundPaths = [
					// hojta "sounds" folder
					path.resolve(__dirname + '\\..\\sounds\\' + fileName),
					// hojta plugin root folder
					path.dirname(pluginOptions.filePath) + '\\' + fileName,
					// VoiceAttack sound folder
					pluginOptions.voiceAttackSoundsPath + '\\' + fileName
				];

				check(soundPaths);
			};
		}

		var helpers = {
			'playSound': playSound
		};

		return function(pluginOptions) {
			var customizedHelpers = {};
			Object.keys(helpers).forEach(function(key) {
				customizedHelpers[key] = helpers[key](pluginOptions);
			});
			return customizedHelpers;
		};
	})();

	fs.watch(__dirname + '/server.js', function() {
		process.exit(0);
	});

	// Setup httpd
	var reqCount = 0;
	var server = http.createServer(function(req, res) {
		if (req.url === '/terminate/') {
			console.log('Got terminated');
			return process.exit(0);
		}

		res.writeHead(200, {
			'Content-Type': 'application/json'
		});

		reqCount += 1;
		var id = reqCount;
		hangingRequests[id] = res;

		var data = '';
		req._hojtaId = id;
		req.on('data', function(chunk) {
			data += chunk;
		});

		req.on('end', function() {
			var pluginData = JSON.parse(data);
			// Don't fiddle with data sent plugin in, just pass it on
			Object.freeze(pluginData);

			var pluginTasks = plugins.map(function(plugin) {
				return plugin(pluginData);
			});

			async.parallel(pluginTasks, sendResponse(res));
		});
	});

	var port = 5922;
	var address = '127.0.0.1';
	server.on('error', function(err) {
		if (err.code === 'EADDRINUSE') {
			require('http').get({
				host: address,
				port: port,
				path: '/terminate/'
			}).on('error', function() {
				server.listen(port, address);
			});
		}
	});

	server.listen(port, address);

	var store = {};

	// Fetch plugins
	var plugins = [];
	var pluginPaths = [];
	dive(__dirname + '/../plugins/', {
		'ignore': 'node_modules'
	}, function(err, file) {
		if (file.match(/\.hojta\.js/)) {
			pluginPaths.push(file);
		}
	}, function() {
		plugins = pluginPaths.map(function(file) {
			return setupPlugin(file);
		});
	});

	// Setup store for plugin and initiate them with data passed
	// from VoiceAttack.
	function setupPlugin(filePath) {
		var plugin = require(filePath);

		// If files/dirs change update the module, good for development
		var watchTimeout;
		fs.watch(filePath, function() {
			if (watchTimeout) {
				clearTimeout(watchTimeout);
			}
			// Add timout since fs.watch has a tendency to trigger change twice
			watchTimeout = setTimeout(function() {
				var fileName = path.basename(filePath);
				var timeStamp = (new Date() + '').match(/..:..:../);
				console.log(timeStamp + '', 'Reloaded', fileName);

				delete require.cache[require.resolve(filePath)];
				plugin = require(filePath);
			}, 10);
		});

		return function(pluginData) {
			return function(done) {
				var shouldRun = plugin.shouldPluginRun(pluginData);
				if (shouldRun) {
					if (!store[filePath]) {
						store[filePath] = {};
					}

					var helpers = helpersGenerator({
						'filePath': filePath,
						'voiceAttackSoundsPath': pluginData.state.VA_SOUNDS
					});

					plugin.run(pluginData, done, store[filePath], helpers);
				} else {
					done(null, {});
				}
			};
		};
	}

	// Take data returned from plugins and format it so it can
	// be sent to parent plugin/VoiceAttack.
	function sendResponse(httpRes) {
		return function(err, pluginResponses) {
			// What is being passed back to parent
			var res = {
				'textValues': {},
				'conditions': {}
			};

			if (err) {
				res.textValues.Error = err;
			}

			// Merge all textValues to one and conditions to one.
			// Things will be overritten if using same key.
			pluginResponses.forEach(function(pluginRes) {
				// If res is an array plugin used respond(textValues, conditions)
				if (Array.isArray(pluginRes)) {
					res.textValues = extend(false, res.textValues, pluginRes[0]);
					res.conditions = extend(false, res.conditions, pluginRes[1]);
				} else {
					res.textValues = extend(false, res.textValues, pluginRes);
				}
			});

			httpRes.end(JSON.stringify(res));

			delete hangingRequests[httpRes._hojtaId];
		};
	}
})();